/*
** $PROJECT: cli.datatype
**
** $VER: dispatch.c 39.4 (24.02.97)
**
** by
**
** Stefan Ruppert , Windthorststrasse 5 , 65439 Floersheim , GERMANY
**
** (C) Copyright 1995-1997
** All Rights Reserved !
**
** $HISTORY:
**
** 24.02.97 : 039.004 : added code to fix gpInput problem. See GM_LAYOUT msg.
** 16.02.97 : 039.003 : added INFILE option
** 13.11.95 : 039.002 : finished v39.2
** 06.08.95 : 039.001 : initial
**
*/

/*FS*//* ------------------------------- include -------------------------------- */

#include "classbase.h"

#include <dos/dostags.h>

/*FE*/

/*FS*//* ------------------------------- autodoc -------------------------------- */

/*GB*** cli.datatype/cli.datatype ********************************************
*
*    NAME
*        cli.datatype - data type for launching extern programs and
*            displaying the command's standard output
*
*    FUNCTION
*        This datatype launches an extern command and displays the standard
*        output of that command. This is useful to view the contents of a
*        archive like lha or tar archives without writing a complete datatype.
*        The standard output is placed in a tempory file in "T:", because the
*        datatypes.library can't handle pipes. From V39.2 the cli.datatype
*        uses NewDTObject() for the tempory file. So it's possible to display
*        files, which have special datatypes (e.g. Unix-Manual-Pages).
*
*    PREFS
*        The cli datatype is configured by a preference file "cli.prefs" in
*        "PROGDIR:Prefs/Datatypes/" or "Env:Datatypes/".
*        The file is parsed using ReadArgs() with the following template :
*
*        "DATATYPE/A,STACK/N/K,SUFFIX/K,INFILE/K,COMMAND/F/K/A"
*
*        DATATYPE - specifies the datatype
*        STACK    - size of stack to use for the command
*        SUFFIX   - suffix to use for tempory file, thus the datatypes.library
*            can determine the type of the file (e.g: ".0" for Man-Datatype)
*        INFILE   - use explicit the specified file instead of the generated
*            stdout tempory file.
*            NOTE: This file will be deleted when the object is disposed.
*        COMMAND  - specifies the command to launch for that datatype, a "%s"
*            in the command string is replaced by the real filename.
*
*    EXAMPLE
*        ; cli.datatype preference file
*        Lha                         COMMAND=lha v %s
*        Tar                         COMMAND=tar tvf %s
*        NRoff STACK=50000 SUFFIX=.0 COMMAND=bin:groff -man -Tascii %s
*
*    METHODS
*        OM_NEW - create a new object and launch the extern command
*
*        OM_DISPOSE - delete the object
*
*    TAGS
*        none
*
*    NOTE
*        With the used method of creating and handling of tempory output, it's
*        also possible to create a special datatype for crunshed files through
*        PowerPacker or XPK. All what have to be done is to create a tempory
*        file and then call NewDTObject() on that. The source of the
*        cli.datatype gives a the skeleton for such datatypes. Please feel free
*        to implement XPK or PowerPacker datatypes. I don't do this.
*
*    SEE ALSO
*        text.datatype, ascii.datatype, datatypes.library
*
******************************************************************************
*
*/

/*FE*/

/*FS*//* ------------------------------- defines -------------------------------- */

#define OPSET(msg)         ((struct opSet *) (msg))
#define OPGET(msg)         ((struct opGet *) (msg))
#define BUFFER_SIZE        1024
#define GAD(x)             ((struct Gadget *) (x))
#define TEMPLATE           "DATATYPE/A,STACK/N/K,SUFFIX/K,INFILE/K,COMMAND/F/K/A"

struct TemplateData
{
   STRPTR td_DataType;
   ULONG *td_Stack;
   STRPTR td_Suffix;
   STRPTR td_InFile;
   STRPTR td_Command;
};

/*FE*/

/*FS*//* ------------------------------ read prefs ------------------------------ */

LONG readPrefs(struct ClassBase *cb,STRPTR file,struct Program *prg)
{
   STRPTR buf = prg->Buffer;
   LONG retval = -1;
   BPTR fh;

   if((fh = Open(file,MODE_OLDFILE)) != NULL)
   {
      struct RDArgs *rdargs;
      struct RDArgs *args;
      /* the real buffer is 2 times of BUFFER_SIZE */
      STRPTR line = buf + BUFFER_SIZE;
      struct TemplateData td = {NULL};

      if((rdargs = (struct RDArgs *) AllocDosObject(DOS_RDARGS,NULL)) != NULL)
      {
	 DB(("rdargs at : %lx\n",rdargs));

	 while(FGets(fh,line,BUFFER_SIZE) != NULL && retval == -1)
	 {
	    /* skip comment and empty lines, to speed up parsing */
	    if(*line == ';' || *line == '\n')
	       continue;

	    rdargs->RDA_Source.CS_Buffer = line;
	    rdargs->RDA_Source.CS_Length = strlen(line);
	    rdargs->RDA_Source.CS_CurChr = 0;

	    if((args = ReadArgs(TEMPLATE,(LONG *) &td,rdargs)) != NULL)
	    {
	       DB(("args at %lx\n",args));

	       if(!Stricmp(prg->Name,td.td_DataType))
	       {
		  if(td.td_Stack)
		     prg->Stack = *td.td_Stack;

		  if(td.td_Suffix)
		     strcpy(prg->Suffix,td.td_Suffix);

		  if(td.td_InFile)
		     strcpy(prg->InFile,td.td_InFile);

		  strcpy(buf,td.td_Command);
		  retval = 0;
	       }

	       FreeArgs(args);
	    }
	 }
	 FreeDosObject(DOS_RDARGS , rdargs);
      }
      Close(fh);
   }

   return(retval);
}
/*FE*/

/*FS*//* --------------------------- class dispatcher --------------------------- */

ClassCall ULONG dispatcher(REGA0 Class *cl,REGA2 Object *obj,REGA1 Msg msg)
{
   struct ClassBase *cb = (struct ClassBase *) cl->cl_UserData;
   struct CliData *cd = INST_DATA(cl,obj);
   ULONG retval = 0;

   switch(msg->MethodID)
   {
   case OM_NEW:
      {
	 struct TagItem *attrs = OPSET(msg)->ops_AttrList;
	 struct TagItem *clone;
	 STRPTR buf;

	 if((clone = CloneTagItems(attrs)) != NULL)
	 {
	    struct Process *myproc = (struct Process *) FindTask(NULL);
	    struct TagItem *tag;
	    LONG rc;

	    BPTR handle = NULL;
	    STRPTR realname = NULL;

	    if((buf = AllocMem(sizeof(struct Program) + 2 * BUFFER_SIZE,MEMF_ANY)) != NULL)
	    {
	       STRPTR cmd  = buf + BUFFER_SIZE;
	       struct Program *prg = (struct Program *) (cmd + BUFFER_SIZE);

	       /* get the real name of the file */
	       if((tag = FindTagItem(DTA_Name,clone)) != NULL)
	       {
		  realname = (STRPTR) tag->ti_Data;
		  tag->ti_Tag = TAG_IGNORE;
	       }

	       if((tag = FindTagItem(DTA_DataType,clone)) != NULL)
	       {
		  struct DataTypeHeader *dth = ((struct DataType *) tag->ti_Data)->dtn_Header;

		  tag->ti_Tag = TAG_IGNORE;

		  prg->Name   = dth->dth_Name;
		  prg->Buffer = buf;
		  prg->Stack  = 4096;
		  prg->InFile[0] = 0;
		  prg->Suffix[0] = 0;

		  /* get tool to launch */
		  if(myproc->pr_HomeDir == NULL || ((rc = readPrefs(cb,"PROGDIR:Prefs/Datatypes/cli.prefs",prg)) == -1))
		     rc = readPrefs(cb,"Env:Datatypes/cli.prefs",prg);

		  /* is there a extern program for that datatype ? */
		  if(rc != -1)
		  {
		     if((tag = FindTagItem(DTA_Handle,clone)) != NULL)
		     {
			handle = (BPTR) tag->ti_Data;
			tag->ti_Tag = TAG_IGNORE;
		     }
		  }
	       }

	       DB(("handle   = 0x%lx\n",handle));
	       DB(("realname = \"%s\"\n",realname));

	       if(handle != NULL && realname != NULL)
	       {
		  STRPTR name;

		  if((name = AllocVec(256,MEMF_ANY)) != NULL)
		  {
		     BPTR pdir;
		     BPTR olddir;

		     if(prg->InFile[0] == '\0')
		     {
			STRPTR file = FilePart(realname);
			STRPTR ptr = file + strlen(file);

			while(ptr >= file)
			{
			   ptr--;
			   if(*ptr == '.')
			   {
			      *ptr = '\0';
			      break;
			   }
			}

			/* create a unique name */
			sprintf(name,"t:.%s%ld%s%s",prg->Name,(LONG)cb->cb_UniqueCnt++,file,prg->Suffix);

			if(ptr >= file)
			   *ptr = '.';

			/* append redirection */
			strcat(buf," >%s");
		     } else
		     {
			strcpy(name,prg->InFile);
		     }

		     /* now format command line to launch */
		     sprintf(cmd,buf,FilePart(realname),name);

		     pdir = ParentDir(handle);
		     olddir = CurrentDir(pdir);

		     DB(("run command : \"%s\"\n",cmd));

		     /* now run the command and get the output into the tempory file */
		     rc = SystemTags(cmd,NP_StackSize,prg->Stack,
					 NP_Name     ,prg->Name,
					 TAG_DONE);

		     CurrentDir(olddir);
		     UnLock(pdir);

		     if(rc == 0)
		     {
			Object *newobj;
			Tag filtertags[] = {DTA_SourceType,DTA_Handle,DTA_DataType,TAG_DONE};

			DB(("before DoSuperMethodA() !\n"));

			FilterTagItems(clone,filtertags,TAGFILTER_NOT);

			DTL(OPSET(msg)->ops_AttrList);

			if((newobj = (Object *) DoSuperMethodA(cl,obj,msg)) != NULL)
			{
			   cd = INST_DATA(cl,newobj);

			   DB(("before NewDTObjectA() !\n"));

			   DTL(clone);

			   if((cd->cd_Object = NewDTObjectA(name,clone)) != NULL)
			   {
			      SetAttrs(newobj,DTA_ObjName,realname,TAG_DONE);

			      DB(("slave obj at 0x%lx\n",cd->cd_Object));

			      DB(("obj at 0x%lx\n",newobj));

			      cd->cd_Name = name;

			      retval = (ULONG) newobj;
			   }

			   if(retval == 0)
			      CoerceMethod(cl,newobj,OM_DISPOSE);
			}
		     }

		     if(retval == 0)
		     {
			DeleteFile(name);
			FreeVec(name);
		     }
		  }
	       }
	       FreeMem(buf,sizeof(struct Program) + 2 * BUFFER_SIZE);
	    }
	    FreeTagItems(clone);
	 }
      }
      break;
   case OM_DISPOSE:
      {
	 if(cd->cd_Object != NULL)
	    DisposeDTObject(cd->cd_Object);

	 if(cd->cd_Name != NULL)
	 {
	    /* delete tempory file */
	    DeleteFile(cd->cd_Name);
	    FreeVec(cd->cd_Name);
	 }

	 retval = DoSuperMethodA(cl,obj,msg);
      }
      break;
   case DTM_PROCLAYOUT:
   case GM_LAYOUT:
      retval = DoMethodA(cd->cd_Object,msg);
      *((struct IBox *) &GAD(obj)->LeftEdge) = *((struct IBox *) &GAD(cd->cd_Object)->LeftEdge);
      GAD(obj)->Flags = GAD(cd->cd_Object)->Flags;
      GAD(obj)->Activation = GAD(cd->cd_Object)->Activation;
      break;
   default:
      retval = DoMethodA(cd->cd_Object,msg);
   }

   return(retval);
}
/*FE*/

